package edu.tuwien.kbs.keywords;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import org.javatuples.Pair;
import org.javatuples.Triplet;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLObjectProperty;

import redis.clients.jedis.Jedis;

import edu.tuwien.kbs.SemSearchCli;
import edu.tuwien.kbs.dl.DLKBLoader;
import edu.tuwien.kbs.search.QueryRewritter;
import edu.tuwien.kbs.search.SemSearch;

public class KeywordGenerator {

    
    //protected static String GeoConceptsABox = "data/GeoConceptsKBS_ABox.owl";
    //protected static String GeoConceptsTBox = "data/GeoConceptsKBS_TBox.owl";
    // protected static String GeoConceptsKB = "data/GeoConceptsKBS_KB.owl";

    protected static String JedisKeyNames = "Kw_Names";
    protected static String JedisKeyMaps = "Kw_Uris";
    protected static String JedisKeyConceptAsserts = "Kw_ConceptAssertions";
    protected static String JedisKeyTransClosGraph = "Kw_TransClosGraph";   
    protected static String JedisKeyConceptInclusions = "Kw_ConceptInclusions";   
    protected static String JedisKeyObjRolesDomainForRange = "Kw_ObjRolesDomainForRange";   
    // protected static char PrefixTermination =  '#';
    private static  char PrefixTermination  = ((char) 31); // Unit Unit separator
    
    protected Jedis mJedis;
    protected DLKBLoader mKBloader;
    protected String mGeoConceptsKB;
    
    public KeywordGenerator() throws Exception {

	mJedis = new Jedis("localhost");
	mKBloader = new DLKBLoader();
	
	try {
	    Long size = mJedis.dbSize();
	} catch (Exception e) {
	   //Exception en = new Exception("Jedis is not running.", e);
	   //throw en; 
	}	
	
	Properties properties = new Properties();
	try {
	  properties.load(new FileInputStream(SemSearchCli.PropertiesFileName));
	  mGeoConceptsKB = properties.getProperty("geokbfile");
	} catch (IOException e) {
	    // e.printStackTrace();
	    Exception en = new Exception("Property file in execution path is missing.", e);
	    throw en; 
	}
	
    }
    
    public void initKeywords() {
	
	clearKeywords();
	// Order also influences order for override of Uris (ABox Kws override TBox Kws
	generateTBoxFullNames(true);
	generateABoxFullNames();
    }
    
    
    public KeywordResults getKeywords(String[] searchStrings, int length, boolean combineKeywords) {

	KeywordResults results = new KeywordResults();

	// Search for existing keywords for search string
	for (int i = 0; i < searchStrings.length; i++) {
	    String searchStr = searchStrings[i];
	    ArrayList<String> solutions = new ArrayList<String>();
	    // Ignore single letter search strings
	    if (searchStr.length() <= 1)
		continue;

	    // Add Keywords from the Onto.
	    solutions = completeKeywords(searchStr, length);
	    
	    // If more possible solution, choose the one that contains also the following keyword
	    if (!combineKeywords && (i + 1) < searchStrings.length) {
		String singleSol = null;
		String tempSearch = searchStrings[i + 1].toLowerCase();
		for (String tempSol : solutions) {
		    if (tempSol.contains(tempSearch)) {
			singleSol = tempSol;
			break;
		    }
		}
		if (singleSol != null) {
		    solutions.clear();
		    solutions.add(singleSol);
		    i++;
		}
	    }

	    results.add(searchStr, solutions);
	}
	
	// Combine single keywords if connected
	if(combineKeywords) 
	    combineKeywords(results);
	
	return results;

    }
    
    
    public void generateABoxFullNames() { // boolean clearExisting
	
	// Load individuals as the tuple <uri, label>
	mKBloader.loadKB( mGeoConceptsKB); // loadABox GeoConceptsABox // System.getProperty( "user.dir" ) + "/" +
	ArrayList<Pair<String,String>> individuals = mKBloader.getABoxIndividualLabels();

	System.out.println("Fill ABox search string:");

	// Fill indiviudals to Redis, always replaces existing values
	FillStringsToRedis(individuals, true ); // false

	// mKBloader.loadTBox(GeoConceptsTBox ); 
	ArrayList<Pair<String,String>> conceptAsserts = mKBloader.getConceptAssertions();
	
	
	 // Read individuals, and assign class to it (consider first assertion)
	for (Pair<String, String> pair : conceptAsserts) {
	    if(! mJedis.hexists(JedisKeyConceptAsserts,pair.getValue0())) {
		mJedis.hset(JedisKeyConceptAsserts, pair.getValue0(), pair.getValue1() );
	    }
	}
    } 

    public void generateTBoxFullNames(boolean log) {
	
	// Load classes as the tuple <uri, label>
	mKBloader.loadKB(mGeoConceptsKB); // loadTBox GeoConceptsTBox
	ArrayList<Pair<String,String>> individuals = mKBloader.getTBoxConceptsRoleLabels();
	
	System.out.println("Fill TBox search strings:");
	
	// Fill individuals to Redis, skips existing values, 
	// this happens if Concept has an Individual with the same label (singleton).
	FillStringsToRedis(individuals, true);
	
	// Get concepts
	Set<OWLClass> concepts = mKBloader.getTBoxConcepts();
	
	// Get and add all ObjRoles and their domain and ranges
	Set<OWLObjectProperty> objRoles = mKBloader.getTBoxObjectRoles();
	
	for (OWLObjectProperty role : objRoles) {
	    
	    ArrayList<String> domains = mKBloader.getDomainsOfObjectRole(role);
	    ArrayList<Pair<String,Integer>> ranges = mKBloader.getRangesOfObjectRole(role);
	    
	    String roleUri = mKBloader.getShortUriForObjectRole(role);
	    	    
	    for (int i = 0; i < ranges.size(); i++) {
		Pair<String,Integer> ran = ranges.get(i);
		for (int j = 0; j < domains.size(); j++) {
		    String domUri = domains.get(j);
		    String role_range = roleUri + ";" + domUri + ";" + ran.getValue1();
		    mJedis.hset(JedisKeyObjRolesDomainForRange, ran.getValue0(), role_range);		    
		}
	    }	    
	    
	}
	
	// Create the (n x n) concept matrix with key concept1_concept2 value 0 or 1. 
	
	// Direct connected (c.f. ESWC'13 paper)
	for (OWLClass cls1 : concepts) {
	    String concept1Uri = SemSearch.DefaultNamespacePrefixTBox + ":" + cls1.getIRI().getFragment();
	    	    
	    // Entirely rewritten according to ESWC'13 paper
	      	    
  	    // Get concepts that are reachable directly through its ObjectRoles
	    HashMap<String, Integer> reachableConcepts = new HashMap<String, Integer>();
	    mKBloader.getConnectedConceptUris(reachableConcepts, cls1);
	    
	    // Get and add concepts that are reachable by a superclass through its ObjectRoles
	    mKBloader.getConnectedConceptToSuperConceptUris(reachableConcepts, cls1);
	    
	    for (String key : reachableConcepts.keySet()) {
		String idUris = concept1Uri + "_" + key;
		mJedis.hincrBy(JedisKeyTransClosGraph, idUris, 1);
		if(log) System.out.println(concept1Uri + " -> " + key);
	    }
	    
	    // Get superclass of a concept
	    HashMap<String, String> superConcepts = new HashMap<String, String>();
	    mKBloader.getSuperConceptUris(superConcepts, cls1);
	    
	    // Get and add all concepts inclusions
	    for (String key : superConcepts.keySet()) {
		String idUris = concept1Uri + "_" + key;
		mJedis.hincrBy(JedisKeyConceptInclusions, idUris, 1);
		if(log) System.out.println(concept1Uri + " -> " + key);
	    }
	}
	
	// Indirect connected (c.f. ESWC'13 paper)
	QueryRewritter rewrt = new QueryRewritter();
	
	for (OWLClass cls1 : concepts) {
	    String concept1Uri = SemSearch.DefaultNamespacePrefixTBox + ":" + cls1.getIRI().getFragment();
	    for (OWLClass cls2 : concepts) {
		    String concept2Uri = SemSearch.DefaultNamespacePrefixTBox + ":" + cls2.getIRI().getFragment();
		    
		    if(cls1 == cls2) continue;
		    
		    Triplet<String, String, String> ret = rewrt.optimizeSpatialFeatureForTwoRoles(this, concept1Uri, concept2Uri);
		    
		    // Check if a witness of the two concepts exists, and if the witness is subsumed/Equal to geot:BuildingFeature
		    if(ret!=null) {
			if (ret.getValue0().equals(QueryRewritter.BaseConceptBuilding) || checkConceptInclusion(ret.getValue0(),QueryRewritter.BaseConceptBuilding)) {
				String idUris = concept1Uri + "_" + concept2Uri;
				mJedis.hincrBy(JedisKeyTransClosGraph, idUris, 1);
				if(log) System.out.println(concept1Uri + " -> " + concept2Uri);
			}
		    }
	    }
	}
	
    } 

    public ArrayList<SimpleConceptRoleLevel >  getSortedDomainObjRolesForRange(String conceptForRange) { // Triplet<String, String, Integer>
	
	List<String> objecRolesDomains =  mJedis.hmget(JedisKeyObjRolesDomainForRange, conceptForRange);
		
	ArrayList<SimpleConceptRoleLevel> results = new ArrayList<SimpleConceptRoleLevel>();

	for(String objRolesDom : objecRolesDomains) {
	    if(objRolesDom == null) continue;
	    String[] vals = objRolesDom.split(";");
	    SimpleConceptRoleLevel res = new SimpleConceptRoleLevel(vals[0], vals[1], Integer.parseInt(vals[2]));
	    results.add(res);
	}
	
	if(results.size() > 0) 
	    Collections.sort( results, new SimpleConceptRoleLevel.LevelComparator() );
	
	return results;
    }
    
	
    public void clearKeywords() {
		
	// Clean Redis
	if(mJedis.exists(JedisKeyNames))  mJedis.del(JedisKeyNames);
	if(mJedis.exists(JedisKeyMaps)) mJedis.del(JedisKeyMaps);
	if(mJedis.exists(JedisKeyTransClosGraph)) mJedis.del(JedisKeyTransClosGraph);
	if(mJedis.exists(JedisKeyConceptAsserts)) mJedis.del(JedisKeyConceptAsserts);
	if(mJedis.exists(JedisKeyConceptInclusions)) mJedis.del(JedisKeyConceptInclusions);
	if(mJedis.exists(JedisKeyObjRolesDomainForRange)) mJedis.del(JedisKeyObjRolesDomainForRange);
	
	
	System.out.println("Keyword index cleared...");
  
    }
    
    public void flushRedis(boolean onlyActualDB) {

	if(onlyActualDB)
	    mJedis.flushDB();
	else
	    mJedis.flushAll();		
    }
    

    public String[] getUrisFromSearchStrings(KeywordResults searchResults) {

	ArrayList<String> results = new ArrayList<String>();
	boolean skipNext = false;
	
	for (int i = 0; i < searchResults.size(); i++) {
	    ArrayList<String> kws = searchResults.getKeywords(i);  
	    if(kws.size() == 0) continue;
	    if(skipNext) {
		skipNext = false;
		continue;
	    }
	    
	    // Use first keyword to map to Uri
	    String kw = kws.get(0);
	    String uri = getUriForLabel(kw);
	    
	    if(uri!=null) {
		results.add(uri);
		// Check next keyword if it is contained in existing, then keywords belong together
		if((i+1) < searchResults.size()) {
		    ArrayList<String> kws2 = searchResults.getKeywords(i+1);
		    if(kws2.size() > 0) {
			if(kw.contains(kws2.get(0)))
			    skipNext = true;
		    }
		}
	    }

	}	
	
	return results.toArray(new String[results.size()]);
	
    }   

    public void combineKeywords(KeywordResults search) {

	ArrayList<KeywordResults> searchGroups = new ArrayList<KeywordResults>();
	// Always one group, if no spatial predicate are entered
	KeywordResults actualGroup = new KeywordResults();
	
	// Build groups of KeywordResults each group separated by a spatial predicate
	for (int i = 0; i < search.size(); i++) {
	    String searchString = search.getSearchString(i);
	    ArrayList<String> singleKws = search.getKeywords(i);
	    boolean isSpatialPredicate = false; 
	    
	    // Check is string is a spatial predicate
	    for (int j = 0; j < singleKws.size(); j++) {
		String kwName = singleKws.get(j); 
		if(kwName == null)  continue;
		//List<String> uris = mJedis.hmget(JedisKeyMaps, kwName);
		String kwUri = mJedis.hmget(JedisKeyMaps, kwName).get(0);

		if(SpatialPredicatesMyITS.hasSpatialPredicate(kwUri)) {
		    // Create new group spatial pred and add last group
		    searchGroups.add(actualGroup);
		    
		    KeywordResults spatialGroup = new KeywordResults(true);
		    spatialGroup.add(searchString, new ArrayList<String>(Arrays.asList(kwName)));
		    searchGroups.add(spatialGroup);
		    
		    actualGroup = new KeywordResults();
		    isSpatialPredicate = true;
		    break;
		}
	    }
	    // add to current group
	    if(!isSpatialPredicate) 
		actualGroup.add(searchString, singleKws);
	    	    
	}
	if(actualGroup!=null && !actualGroup.isSpatialPredicate())
	    searchGroups.add(actualGroup);

	
	// Case without spatial predicats
	if(searchGroups.size()<=1) {
	    combineKeywordsPartly(search);
	}
	// Use normal combination algorithm for each group and combine them
	else {
	    ArrayList<String> combinations = null;
	    for (int i = 0; i < searchGroups.size(); i++) {
		KeywordResults searchGroup = searchGroups.get(i);
		combineKeywordsPartly(searchGroup);
		
		if(searchGroup.getCombinations().length==0) continue;
		
		if(combinations==null) {
		    combinations = new ArrayList<String>(Arrays.asList(searchGroup.getCombinations()));
		}
		else {
		    for (int j = 0; j < combinations.size(); j++) {
			String tempVal = combinations.get(j);			
			combinations.set(j, tempVal + " " + searchGroup.getCombinations()[0]);
		    }
		}
	    }	
	    
	    search.setCombinations(combinations);
	    
	}
	
	
    }
    
    
    public void combineKeywordsPartly(KeywordResults search) {

	ArrayList<String> singleKwsTotal = new ArrayList<String>();
	// ArrayList<String> singleKwsCombined = new ArrayList<String>();
	// HashMap<String, Integer> addedKwds = new HashMap<String, Integer>();
	boolean foundExtension = false;
	ArrayList<String> singleKwsTemp = new ArrayList<String>();
	
	for (int i = 0; i < search.size(); i++) {
	    ArrayList<String> singleKws = search.getKeywords(i); 
	    String kwFirstName = null;
	    
	    for (int j = 0; j < singleKws.size(); j++) {
		String kwName = singleKws.get(j); 
		if(kwName == null)  continue;
		
		// Store all combination for a selected kwName and add combinations 
		ArrayList<String> kwdExtensions = new ArrayList<String>();
		// ArrayList<String> tempKwdsCombs = new ArrayList<String>(); 
		
		// Safe first keyword for search string
		if(kwFirstName == null)  kwFirstName = kwName;
		
		// Get Uri for first KW
		//List<String> uris = mJedis.hmget(JedisKeyMaps, kwName);
		String kwUri = mJedis.hmget(JedisKeyMaps, kwName).get(0);
		
		// If the keyword belongs to an indivual, get asserted concept of it. 
		// Reason, the trans. closure of onto graph is just build only from the TBox
		if(kwUri.contains(SemSearch.DefaultNamespacePrefixABox)) {
		    kwUri = getAssertedConept(kwUri);
		}

		// Create valid combination with other keywords. Valid, if connected in the ontology
		for (int k = 0; k < search.size(); k++) {
		    ArrayList<String>  singleKws2 = search.getKeywords(k); 
		    if (i >= k) 
			continue;		    
		    
		    for (int l = 0; l < singleKws2.size(); l++) {
			String kwName2 = singleKws2.get(l); 
			if(kwName2 == null) 
			    continue;
			
			// Get Uri for second kw
			List<String> uris2 = mJedis.hmget(JedisKeyMaps, kwName2);
			String kwUri2 =  getAssertedConept(uris2.get(0));
			
			// Check in adjacency matrix of graph if they is connected
			if (mJedis.hexists(JedisKeyTransClosGraph, kwUri + "_" + kwUri2)) {
			    //tempKwdsCombs.add(kwName + " " + kwName2);
			    kwdExtensions.add(kwName2);
			    // Keep track of the combined keywords, as they do not need to be there as a single keyword
			    //singleKwsCombined.add(kwName);
			    //singleKwsCombined.add(kwName2);
			    //if (log) System.out.println(kwName + " " + kwName2);
			}
			
			// TODO: Check in adjacency matrix of graph if reverse egde is connected
		    }
		}
		
		// Add unary KW
		singleKwsTotal.add(kwName);

		
		//  New use a lattice and check each level
		if(kwdExtensions.size() > 0) { // 1
		    String unfoldedCombs = "";
		    for (int l = 0; l < kwdExtensions.size(); l++) {
			String kwExt = kwdExtensions.get(l); 
			unfoldedCombs = unfoldedCombs + " " + kwExt;
			if(l==3) 
			    break; // Set a limit how long the combinations can be, otherwise it gets to confusing for the user
		    }
		    search.addKWCombination(kwName + " " + unfoldedCombs);
		    foundExtension = true;
		}
		
/*		Pre ESWC'13 Paper
  		// Above combination are all binary, unfold them to create more possible. 
		if(kwdExtensions.size() > 1) {
		    String unfoldedCombs = "";
		    for (int l = 0; l < kwdExtensions.size(); l++) {
			String kwExt = kwdExtensions.get(l); 
			unfoldedCombs = unfoldedCombs + " " + kwExt;
			if(l==3) 
			    break; // Set a limit how long the combinations can be, otherwise it gets to confusing for the user
		    }
		    search.addKWCombination(kwName + " " + unfoldedCombs);
		    
		}
		
		// Add valid binary combination
		search.addKWCombinations(tempKwdsCombs);		
*/		
	    }	

	    
	    if(kwFirstName != null) 
		singleKwsTemp.add(kwFirstName);

	}

	// Add single keyword, if it does not exits in combination
	for (int j = 0; j < singleKwsTotal.size(); j++) {
	    search.addKWCombination(singleKwsTotal.get(j));

	}

	
/*	Pre ESWC'13 Paper
  	if(singleKwsCombined.size()>0) {
	    // Remove combined
	    singleKwsTotal.removeAll(singleKwsCombined);
	    // Add single keyword, if it does not exits in combination
	    for (int j = 0; j < singleKwsTotal.size(); j++) {
		search.addKWCombination(singleKwsTotal.get(j));
	    }
	}
	else {
}*/
	if(!foundExtension && search.size() > 1) {
	    // This is a not-so-good workaround, if keywords do not combine, 
	    // merge for each search string the first keyword together.
	    // Better would be merging combined keywords with non-combined.
		// Above combination are all binary, unfold them to create more possible.
		if(singleKwsTemp.size() > 1) { // singleKwsTotal
		    String unfoldedCombs = "";
		    for (int j = 0; j < singleKwsTemp.size(); j++) {
			String kwExt = singleKwsTemp.get(j); 
			if(j>0) unfoldedCombs = unfoldedCombs + " ";
			unfoldedCombs = unfoldedCombs + kwExt;
		    }
		    search.addKWCombination(unfoldedCombs);
		}

		// Final add the unary keywords as well
		for (int j = 0; j < singleKwsTotal.size(); j++) {
		    search.addKWCombination(singleKwsTotal.get(j));
		}
	}
	//}
   }    
   
    public boolean checkConceptInclusion(String subConcept, String superConcept) { 
	return mJedis.hexists(JedisKeyConceptInclusions, subConcept + "_" + superConcept);	
    }
       
    
    public String getUriForLabel(String label) {
	
	if(! mJedis.hexists(JedisKeyMaps,label)) return null;
	
	//List<String> uris = mJedis.hmget(JedisKeyMaps, label);
	return mJedis.hmget(JedisKeyMaps, label).get(0);
    }

    public  String getAssertedConept(String indivURI) {
	
	    if(mJedis.hexists(JedisKeyConceptAsserts, indivURI)) {
		List<String> uris3 = mJedis.hmget(JedisKeyConceptAsserts, indivURI);
	    	return uris3.get(0);
	    }
	    else {
		// if no assertion return indiv.
		return indivURI;
	    }
    }
     
    protected  ArrayList<String> completeKeywords(String prefix, int length) { //  String[] 
	    
	ArrayList<String> results = new ArrayList<String>();
	
	String lcPrefix = prefix.toLowerCase();
	int rangelen = 50  ; // This is not random, try to get replies < MTU size
	int resCount = 0;
	
	Long temp = mJedis.zrank(JedisKeyNames, lcPrefix);
	if(temp==null) return results;

	int start = temp.intValue();
	
	while (resCount != length) {
	    Set<String> ranges = mJedis.zrange(JedisKeyNames, start, start+rangelen-1);
	    start += rangelen;
	    
	    if(ranges.size() == 0) break;
	    
	    Iterator<String> itr = ranges.iterator(); 
	    while(itr.hasNext()) {
		String val = itr.next();
		
		// Stop condition if entry different to prefix is reached
		int minVal = Math.min(lcPrefix.length(), val.length());
		if(! val.regionMatches(0, lcPrefix, 0, minVal)) break;
		
		// if entry is marked with *, complete keyword
		if(val.charAt(val.length()-1) == PrefixTermination && resCount != length)  {
		    String name = val.substring(0, val.length()-1);
		    results.add(name); // + " (" + uris.get(0) +")";
		    resCount++;
		}
	    }
	    
	}
	
	return results;
		
    }

    
    protected void generateNamesinJedis(String name, boolean log) {
	
	// Generate in lexicographic order all substrings from left to right (c, ca, caf, cafe, cafe#) and
	// mark full name with #.
	for (int j = 0; j < (name.length()); j++) {
	    String pref = name.substring(0, j + 1);
	    mJedis.zadd(JedisKeyNames, 0, pref);
	    if (log)
		System.out.println(pref);
	}
	
	// Just terminate names larger 1
	if(name.length() > 1) {
	    mJedis.zadd(JedisKeyNames, 0, name + PrefixTermination);
	    if (log) System.out.println(name + PrefixTermination);
	}

    }

    protected void FillStringsToRedis(ArrayList<Pair<String,String>> pairs, boolean skipIfExists) {
	
	for (Pair<String, String> pair : pairs) {
	    
	    String label =  pair.getValue1().toLowerCase();
	    
	    // if key already exists, skip the generation
	    if(skipIfExists && mJedis.hexists(JedisKeyMaps,label)) continue;
	    
	    // Generate Names and all possible prefixes, (e.g. cafe; caf; ca; c) and mark valid with *
	    generateNamesinJedis(label, true);
	    
	    // Keep for every name the Uri
	    mJedis.hset(JedisKeyMaps, label.toLowerCase(), pair.getValue0() );
	    
	} 	
    }
    
       
 
    
}
