package edu.tuwien.kbs.ldm.parser;



/**
 * Represents a literal in the rule. A literal consists of a predicate, a pattern representing
 * the variable names or constants and a flag determining whether the literal is negated.
 */
public class Literal implements ILiteral {

    protected Predicate m_predicate;
    protected boolean m_isPositive; /** Set to <code>true</code> if this literal is positive under NAF (Classical neg. is in Prediacte). */
    protected ITerm[] m_terms;
    protected int m_hashCode;
    protected boolean m_isRecursive; 

    protected String mAdditionalParameter;
    
    /**
     * Creates a literal.
     *
     * @param predicate                         the predicate
     * @param isPositive                        <code>true</code> if the literal is positive
     * @param terms                             the terms of the literal
     */
    public Literal(Predicate predicate,boolean isPositive,ITerm[] terms) {
        this(predicate,isPositive,terms,false);
    }
    
    /**
     * Creates a literal with an additional variable for selecting the information about a disjunction.
     *
     * @param predicate                         the predicate
     * @param isPositive                        <code>true</code> if the literal is positive
     * @param terms                             the terms of the literal
     * @param hasDisjunctionVariable            <code>true</code> if the literal is disjunctive
     */
    public Literal(Predicate predicate,boolean isPositive,ITerm[] terms,boolean hasDisjunctionVariable) {
        m_predicate=predicate;
        m_isPositive=isPositive;
        m_terms=terms;
        
        if (m_terms.length!=m_predicate.getArity()+(hasDisjunctionVariable ? 1 : 0))
            throw new IllegalArgumentException("Incorrect number of terms in the literal.");
        
        m_hashCode=m_predicate.hashCode()*3+(m_isPositive ? 1 : 0);
        
        for (int i=0;i<m_terms.length;i++)
            m_hashCode=m_hashCode*7+m_terms[i].hashCode();
    }

    public String getEntityAttribueNames() 	{

	StringBuffer argFields = new StringBuffer();    		
	for (int i = 0; i < m_terms.length; i++) {
	    if(i>0) argFields.append(", ");
	    argFields.append(ConstantsDSR.TABLE_ATTRIBUTE_NAME + (i+1));
	}	

	return argFields.toString();
    }

    public boolean IsAggregate() {
	return (mAdditionalParameter!=null && mAdditionalParameter.length() > 0);
    }
    
    public String getAddParameter() {
	return mAdditionalParameter;
    }
    public void setAddParameter(String para) {
	mAdditionalParameter = para;
    }

    public boolean IsRecursive() {
	return m_isRecursive;
    }
    public void setIsRecursive(boolean rec) {
	m_isRecursive = rec;
    }

    public Predicate getPredicate() {
	return m_predicate;
    }
    
    public String getPredicateName() {
	return m_predicate.getName();
    }
    
    public Predicate[] getEmbeddedPredicates() {
		return null;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getPredicateFullName()
	 */
    public String getPredicatNameAsView() {
        return m_predicate.getNameAsView();
    } 
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#isPositive()
	 */
    public boolean isPositive() {
        return m_isPositive;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getArity()
	 */
    public int getArity() {
        return m_terms.length;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#isArgumentBoundToVariable(int)
	 */
    public boolean isArgumentBoundToVariable(int argumentIndex) {
        return m_terms[argumentIndex] instanceof Variable;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#isArgumentBoundToConstant(int)
	 */
    public boolean isArgumentBoundToConstant(int argumentIndex) {
        return m_terms[argumentIndex] instanceof Constant;
    }

    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getTerm(int)
	 */
    public ITerm getTerm(int argumentIndex) {
        return m_terms[argumentIndex];
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getArgumentVariable(int)
	 */
    public Variable getArgumentVariable(int argumentIndex) {
        ITerm term=m_terms[argumentIndex];
        if (term instanceof Variable)
            return (Variable)term;
        else
            return null;
    }
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getArgumentConstant(int)
	 */
    public Constant getArgumentConstant(int argumentIndex) {
        ITerm term=m_terms[argumentIndex];
        if (term instanceof Constant)
            return (Constant)term;
        else
            return null;
    }
    
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#getTerms()
	 */
    public ITerm[] getTerms() {
        return m_terms;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#containsVariable(edu.tuwien.ps.datalog.Variable)
	 */
    public boolean containsVariable(Variable variable) {
        for (int i=m_terms.length-1;i>=0;--i)
            if (m_terms[i].equals(variable))
                return true;
        return false;
    }
    
    /* (non-Javadoc)
	 * @see edu.tuwien.ps.datalog.ILiteral#positionOfVariable(edu.tuwien.ps.datalog.Variable)
	 */
    public int positionOfVariable(Variable variable) {
        for (int i = 0; i < m_terms.length; i++) {
            if (m_terms[i].equals(variable))
                return i;
        }
        return -1;
    }
    
    /**
     * Converts this literal to a string.
     */
    public String toString() {
        StringBuffer buffer=new StringBuffer();
        if (!m_isPositive)
            buffer.append("not "); // "~"
        buffer.append(m_predicate.getSimpleName());
        int arity=m_predicate.getArity();
        if (arity>0) {
            buffer.append("(");
            for (int i=0;i<arity;i++) {
                if (i>0)
                    buffer.append(",");
                buffer.append(m_terms[i].toString());
            }
            buffer.append(")");
        }
        return buffer.toString();
    }
    
    /**
     * Compares this object to another.
     */
    public boolean equals(Object that) {
        if (this==that)
            return true;
        
        if (!(that instanceof Literal))
            return false;
        
        Literal thatLiteral=(Literal)that;
        if (!m_predicate.equals(thatLiteral.getPredicate()) || m_isPositive!=thatLiteral.m_isPositive || m_terms.length!=thatLiteral.m_terms.length)
            return false;
        for (int i=m_terms.length-1;i>=0;--i)
            if (!m_terms[i].equals(thatLiteral.m_terms[i]))
                return false;
        return true;
    }
    
    /**
     * Returns the hash-code of this object.
     */
    public int hashCode() {
        return m_hashCode;
    }



    
}
