package edu.tuwien.kbs.ldm.parser;


/**
 * Represents a predicate in the program.
 */
public class Predicate {
    /** Set to <code>true</code> if this predicate is positive under classical negation (NAF is placed in Predicate. */
    protected boolean m_isPositive;
    protected String m_name;
    protected int m_arity;
    protected String m_fullName;  /** The full name of the predicate which includes its positive/negative status and the arity. */
    protected String m_sqlName;
    
    /**
     * Creates a predicate.
     *
     * @param isPositive                <code>true</code> if this predicate is positive under classical negation
     * @param name                      the name of the predicate (excludes the positive/negative status and the arity of the predicate)
     * @param arity                     the arity of the predicate
     */
    public Predicate(boolean isPositive,String name,int arity) {
        m_isPositive=isPositive;
        m_name=name;
        m_arity=arity;
        createFullName();
    }
    
    /**
     * Returns whether this predicate is classically positive.
     */
    public boolean isPositive() {
        return m_isPositive;
    }
    
    /**
     * Returns the name of the predicate (excluding the arity of the predicate and its positive/negative status).
     */
    public String getName() {
        return m_name;
    }
    public String getNameAsView() {
        return ConstantsDSR.SQL_BASE_VIEW_NAME + m_name;
    }
 
    /**
     * Returns the name of the predicate used in an SQL statement. This name can vary from the original name.
     */
    public String getSQLEntityName() {
        if(m_sqlName!=null) return m_sqlName;
    	return m_name;
    }
    
    public void  setSQLEntityName(String name) {
    	m_sqlName = name;
    }
    
    private void createFullName()
    {
        m_fullName=(m_isPositive ? "" : "-")+m_name+"/"+m_arity; 
    }
    
    /**
     * Returns the simple name of the predicate (excluding the arity of the predicate but including its positive/negative status).
     */
    public String getSimpleName() {
        return (m_isPositive ? "" : "-")+m_name;
    }
    
    /**
     * Returns the extended attribute name in the DB (t1.att1, t1.att2, ...)
     */
	public String createSQLTableAttribute(int index, boolean useViews)
	{   
		return createSQLTableAttribute(index, useViews, true);
	}
	
	public String createSQLTableAttribute(int index, boolean useViews, boolean includeEntity)
	{
		StringBuffer buf = new StringBuffer();
		if(includeEntity) {
			if(useViews) buf.append( ConstantsDSR.SQL_BASE_VIEW_NAME);
			buf.append( this.getSQLEntityName() );
			buf.append( "." );
		}
		buf.append( ConstantsDSR.TABLE_ATTRIBUTE_NAME +(index+1) ); 
		return buf.toString();
	}
    
	
    /**
     * Returns the simple attribute name in the DB (att1, att2, ...)
     */
	public String createSQLSimpleTableAttribute( int index)
	{
		return ConstantsDSR.TABLE_ATTRIBUTE_NAME + (index+1);
	}
    /**
     * Returns the full name of the predicate (including the arity of the predicate).
     */
    public String getFullName() {
        return m_fullName;
    }
    
    /**
     * Returns the arity of the predicate.
     */
    public int getArity() {
        return m_arity;
    }
    
    /**
     * Converts this predicate to string.
     */
    public String toString() {
        return m_fullName;
    }
    
    /**
     * Compares this object to another.
     *
     */
    public boolean equals(Object that) {
        if (this==that)
            return true;
        if (!(that instanceof Predicate))
            return false;
        return m_fullName.equals(((Predicate)that).m_fullName);
    }
    
    /**
     * Returns the hash-code of this predicate.
     */
    public int hashCode() {
        return m_fullName.hashCode();
    }
}
