package edu.tuwien.kbs.ldm.parser;

import java.util.ArrayList;

/**
 * Represents a rule in the program. A rule consists of a positive head literal and of a conjunction of
 * several body literals. A rule must be range-restricted (also called safe or allowed), which means that
 * each variable from the head must occur in some positive literal in the body.
 */
public class Rule {

    protected String m_label;
    protected ILiteral[] m_head;
    protected ILiteral[] m_body;
    protected ArrayList<Rule> m_headSiblings; 
    protected boolean m_isFact; 
    protected boolean m_isImported; 
    protected boolean m_isRecursive; 
    protected boolean m_isHeadSiblingRecursive; 

    public int index = -1;
    public int lowlink = -1;
    
	/**
     * Creates an instance of this class.
     *
     * @param label                         the label of the rule
     * @param head                          the head of the rule (must be a positive literal)
     * @param body                          the body of the rule
     */
    public Rule(String label,ILiteral head,ILiteral[] body) {
        this(label,new ILiteral[] { head },body);
    }
    
    /**
     * Creates an instance of this class.
     *
     * @param label                         the label of the rule
     * @param head                          the head of the rule (must be positive literals)
     * @param body                          the body of the rule
     */
    public Rule(String label,ILiteral[] head,ILiteral[] body) {
        m_label=label;
        m_head=head;
        m_body=body;
        
        for (int i=0; i < m_head.length; i++)
            if (!m_head[i].isPositive())
                throw new IllegalArgumentException("Head of the rule must be a positive literal.");
        
        m_headSiblings = new ArrayList<Rule>();
        

    }
    
    /**
     * Returns the label of the rule.
     */
    public String getLabel() {
        return m_label;
    }
 
    /**
     * Returns if this rule has a recursion (head and body have the same predicate).
     */   
    public boolean IsRecursive() {
	return m_isRecursive;
    }
	
    /**
     * Returns if this rule has a recursion (head and body have the same predicate).
     */   
    public boolean hasAggregate() {
	
	for( ILiteral lit : m_body) {
	    
	    if(lit.IsAggregate()) 
		return true;
	}
	
	return false;
	
    }    
    
    /**
     * Returns if a sibling of this rule has a recursion.
     */   
    public boolean IsHeadSiblingRecursive() {
		return m_isHeadSiblingRecursive;
	}
	public void setIsHeadSiblingRecursive(boolean rec) {
		m_isHeadSiblingRecursive = rec;
	}
	
    /**
     * Returns if this rule is a fact.
     */   
    public boolean IsFact() {
		return m_isFact;
	}
    public void setIsFact(boolean fact) {
		m_isFact = fact;
	}
    
    /**
     * Returns if this rule is imported.
     */   
    public boolean IsImported() {
		return m_isImported;
	}
    public void setIsImported(boolean imp) {
		m_isImported = imp;
	}
	
    /**
     * Returns the label of the rule.
     */
    public  ArrayList<Rule> getHeadSiblingList() {
        return m_headSiblings;
    }  
    
    /**
     * Returns the head literal. This method can be called only for Horn rules, otherwise it throws an exception.
     */
    public ILiteral getHeadLiteral() {
        if (m_head.length==1)
            return m_head[0];
        else
            throw new IllegalStateException("This rule doesn't have exactly one head literal.");
    }
    /**
     * Returns the number of head literals.
     */
    public int getHeadLength() {
        return m_head.length;
    }
    /**
     * Returns the head literal with given index.
    */
    public ILiteral getHeadLiteral(int headLiteralIndex) {
        return m_head[headLiteralIndex];
    }
    /**
     * Returns the head literals of the rule.

     */
    public ILiteral[] getHeadLiterals() {
        return m_head;
    }
    /**
     * Returns the number of body literals.
        */
    public int getBodyLength() {
        return m_body.length;
    }
    /**
     * Returns the body literal with given index.
     */
    public ILiteral getBodyLiteral(int bodyLiteralIndex) {
        return m_body[bodyLiteralIndex];
    }
    /**
     * Returns the array of literals of the rule.
     */
    public ILiteral[] getBodyLiterals() {
        return m_body;
    }
    /**
     * Returns <code>true</code> if this is a Horn rule with one head literal.
     */
    public boolean isHorn() {
        return m_head.length==1;
    }
 
 
    /**
     * Returns the string representation of this rule.
     *
     * @return                              the string representation of this rule
     */
    public String toString() {
        StringBuffer buffer=new StringBuffer();
        if (m_label!=null) {
            buffer.append(m_label);
            buffer.append(": ");
        }
        for (int i=0;i<m_head.length;i++) {
            if (i>0)
                buffer.append(" v ");
            buffer.append(m_head[i].toString());
        }
        if (m_body.length!=0) {
            buffer.append(" :- ");
            for (int i=0;i<m_body.length;i++) {
                if (i>0) buffer.append(", ");
                if(m_body[i]!=null) buffer.append(m_body[i].toString());
            }
        }
        buffer.append('.');
        return buffer.toString();
    }
}
